/* $Id$ */
/* Copyright (C) 2006-2008 Adam Wysocki, ArcaBit Sp. z o.o. */

#ifndef ARCADAPI_H__
#define ARCADAPI_H__

#ifdef __cplusplus
extern "C" {
#endif

/**
 * \brief	Scanning mode
 */

enum EAVMode
{
	MODE_NORMAL = 0,	/**< Normal scan, stop after one infection found */
	MODE_RAW,		/**< Raw scan, like normal, but do not recurse archives */
	MODE_CONT		/**< Continuous scan, like normal, but do not stop after one infection found */
};

/**
 * \brief	Answer for the scanner
 */

enum EAVAnswer {
	ANSWER_LEAVE = 0,	/**< Leave with infection */
	ANSWER_CURE,		/**< Try to cure */
	ANSWER_DELETE,		/**< Delete file */
	ANSWER_RENAME,		/**< Rename file to .vir extension */
	ANSWER_BREAK		/**< Stop scanning */
};

/**
 * \brief	One token from the scanner
 *
 * Please see README.arcad-protocol for explanation.
 */

struct SAVResultToken
{
	char *av_key;		/**< Entry name */
	char *av_value;		/**< Value for that entry */
};

/**
 * \brief	Scanning result
 *
 * Scanning result, consists of multiple tokens. Please see 
 * README.arcad-protocol for explanation. 
 */

struct SAVResult
{
	unsigned int av_count;	/**< Number of tokens */
	struct SAVResultToken *av_tokens;	/**< Array of tokens */
};

/**
 * \brief	Constructor
 *
 * Initializes internal structures.
 *
 * \param	verbose	When nonzero, be verbose
 * \return	Session pointer that must be passed to other methods or NULL
 */

void *AVCreate(int verbose);

/**
 * \brief	Destructor
 *
 * Destroys internal structures
 *
 * \param	av	Session pointer
 */

void AVDestroy(void *av);

/**
 * \brief	Get daemon version
 *
 * \param	av	Session pointer
 * \param	buf	Pointer to char buffer
 * \param	bufsz	Size of buffer
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVVersion(void *av, char *buf, int bufsz);

/**
 * \brief	Connect to daemon
 *
 * Connects to daemon and optionally reloads databases.
 *
 * \param	av	Session pointer
 * \param	path	Path to daemon socket
 * \param	reload	When nonzero, reload databases
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVConnect(void *av, const char *path, int reload);

/**
 * \brief	Disconnect from daemon
 *
 * Closes connection to daemon.
 *
 * \param	av	Session pointer
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVDisconnect(void *av);

/**
 * \brief	Pass configuration
 *
 * Passes configuration string to daemon. Intented for internal ArcaBit usage.
 *
 * \param	av	Session pointer
 * \param	config	Configuration string
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVConfig(void *av, const char *config);

/**
 * \brief	Scan file
 *
 * Scans file or directory using given mode.
 *
 * \param	av	Session pointer
 * \param	path	Path to file or directory
 * \param	mode	Scanning mode
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVScanFile(void *av, const char *path, enum EAVMode mode);

/**
 * \brief	Scan stream
 *
 * Scans TCP stream using normal scanning mode.
 *
 * \param	av	Session pointer
 * \return	Daemon port or 0 (error)
 */

int AVScanStream(void *av);

/**
 * \brief	Read result
 *
 * Reads scanning result from daemon. Resulting tokens are sorted lexicoghraphically 
 * by their keys.
 *
 * \param	av	Session pointer
 * \param	result	Result structure pointer
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVResult(void *av, struct SAVResult *result);

/**
 * \brief	Read result token
 *
 * Finds result token using binary search and returns it's value. Passed result tokens 
 * have to be sorted lexicographically by their keys.
 *
 * \param	result	Result structure pointer
 * \param	key	Token name
 *
 * \return	Token value or NULL (not found or error)
 */

const char *AVResultValue(const struct SAVResult *result, const char *key);

/**
 * \brief	Destroy result
 *
 * Destroys SAVResult structure obtained by AVResult. Structure storage 
 * is not freed.
 *
 * \param	result	Result structure pointer
 */

void AVResultDestroy(struct SAVResult *result);

/**
 * \brief	Pass answer
 *
 * Passes answer to daemon, called after reading token "type:ask".
 *
 * \param	av	Session pointer
 * \param	answer	Answer for daemon
 *
 * \retval	0	All ok
 * \retval	-1	Error
 */

int AVAnswer(void *av, enum EAVAnswer answer);

#ifdef __cplusplus
}
#endif

#endif
