/* $Id$ */
/* Copyright (C) 2006-2008 Adam Wysocki, ArcaBit Sp. z o.o. */

#include <arcadapi.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>

#define SOCKET_PATH	"/var/run/arcad.ctl"

#define die(...) do { \
	int se = errno; \
	fprintf(stderr, "arcad-scan: "); \
	errno = se; \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n"); \
	exit(-1); \
} while (0)

extern int errno;

extern char *optarg;
extern int optind, opterr, optopt;

/**
 * \brief	Option structure
 *
 * Structure holding command line options.
 */

struct SOptions
{
	enum EAVMode mode;		/**< Scan mode */
	int verbose;			/**< Be verbose */
	int reload;			/**< Reload databases */
	const char *socket_path;	/**< Path to socket */
	const char *config;		/**< Configuration string */
	char * const *paths;		/**< Paths to scan */
};

/**
 * \brief	Return value from commandline parser
 */

enum EParseRet
{
	PR_OK = 0,	/**< Ok, parsed */
	PR_EXIT_OK,	/**< Parsed, but exit (fast help issued) */
	PR_EXIT_FAIL	/**< Parse error */
};

/**
 * \brief	Fast help
 *
 * Prints fast help on stdout.
 */

static void FastHelp(void)
{
	static const char msg[] = 
		"arcad-scan - arcad command line interface utilizing arcadapi\n"
		"Copyright (C) 2006-2008 ArcaBit Sp. z o.o.\n"
		"\n"
		"Usage: arcad-scan [-hvr] [-m mode] [-s socket path] [-c scanner config] [paths]\n"
		"  -h: This help\n"
		"  -v: Verbose execution\n"
		"  -r: Reload bases\n"
		"\n"
		"Scanning mode: normal, raw, cont\n"
		"\n"
		"Report problems to support@arcabit.com";

	puts(msg);
}

/**
 * \brief	Command line parser
 *
 * Parses command line using given argument list.
 *
 * \param	options	Options structure pointer
 * \param	argc	Options count
 * \param	argv	Options list
 *
 * \retval	PR_OK	Ok, parsed
 * \retval	PR_EXIT_OK	Parsed, but exit (fast help issued)
 * \retval	PR_EXIT_FAIL	Parse error
 */

static enum EParseRet ParseOptions(struct SOptions *options, int argc, char * const argv[])
{
	static const char opts[] = "hm:vrs:c:";
	int done = 0;

	options->mode = MODE_NORMAL;
	options->verbose = 0;
	options->reload = 0;
	options->socket_path = SOCKET_PATH;
	options->config = NULL;
	options->paths = NULL;

	while (!done) {
		switch (getopt(argc, argv, opts)) {
			case 'h':
				FastHelp();
				return PR_EXIT_OK;

			case 'm':
				switch (*optarg) {
					case 'n':
						options->mode = MODE_NORMAL;
						break;

					case 'r':
						options->mode = MODE_RAW;
						break;

					case 'c':
						options->mode = MODE_CONT;
						break;

					default:
						FastHelp();
						return PR_EXIT_FAIL;
				}
				break;

			case 'v':
				options->verbose = 1;
				break;

			case 'r':
				options->reload = 1;
				break;

			case 's':
				options->socket_path = optarg;
				break;

			case 'c':
				options->config = optarg;
				break;

			case -1:
				done = 1;
				break;

			case ':':
			/* FALLTHROUGH */
			case '?':
			/* FALLTHROUGH */
			default:
				FastHelp();
				return PR_EXIT_FAIL;
		}
	}

	options->paths = argv + optind;
	return PR_OK;
}

/**
 * \brief	Find value for given key
 *
 * Searches for given token and returns it's value. Basically wraps AVResultValue() 
 * but guarantees that for mandatory keys the value will be returned (if the key 
 * is not found, will bail out).
 *
 * \param	res	Result structure pointer
 * \param	key	Token key
 * \param	mandatory	If this key is mandatory
 *
 * \retval	Token value or NULL (when token not found and mandatory == 0)
 */

static const char *FindValue(const struct SAVResult *res, const char *key, int mandatory)
{
	const char *value;

	if (!(value = AVResultValue(res, key)) && mandatory)
		die("Cannot find value for key \"%s\"", key);

	return value;
}

/**
 * \brief	Generic event handler
 *
 * Event handler for all scanner responses. Basically prints out keys and their 
 * values for all tokens. Keys and values are unescaped.
 *
 * \param	res	Result structure pointer
 * \param	...	{char *key, int mandatory} pairs, last key == NULL
 */

static void HandleGeneric(const struct SAVResult *res, ...)
{
	int first = 1;

	printf("Result: ");

	va_list ap;
	va_start(ap, res);

	for (;;) {
		const char *key;
		const char *value;
		int mandatory;

		key = va_arg(ap, char *);
		mandatory = va_arg(ap, int);

		if (!key)
			break;

		value = FindValue(res, key, mandatory);
		if (value) {
			printf("%s%s=\"%s\"", first ? "" : ", ", key, value);
			first = 0;
		}
	}

	va_end(ap);
	printf("\n");
}

/**
 * \brief	"scanner" type replies handler
 *
 * Wraps HandleGeneric() with scanner-specific token list.
 *
 * \param	av	Session pointer
 * \param	res	Result structure pointer
 * \param	end	Ending flag pointer
 *
 * \retval	0	Not infected or infection status unknown
 * \retval	!0	Object is infected
 */

static int HandleScanner(void *av, const struct SAVResult *res, int *end)
{
	const char *clean;

	HandleGeneric(res, 
		"type", 1,
		"code", 1,
		"clean", 0,
		NULL, 0);

	*end = !strcmp(FindValue(res, "code", 1), "end");

	clean = FindValue(res, "clean", 0);
	return clean && !strcmp(clean, "no");
}

/**
 * \brief	"create" type replies handler
 *
 * Wraps HandleGeneric() with create-specific token list.
 *
 * \param	av	Session pointer
 * \param	res	Result structure pointer
 */

static void HandleCreate(void *av, const struct SAVResult *res)
{
	HandleGeneric(res, 
		"type", 1,
		"file", 1,
		"object", 1,
		"action", 1,
		"status", 1,
		"progress", 0,
		NULL, 0);
}

/**
 * \brief	"parse" type replies handler
 *
 * Wraps HandleGeneric() with parse-specific token list.
 *
 * \param	av	Session pointer
 * \param	res	Result structure pointer
 */

static void HandleParse(void *av, const struct SAVResult *res)
{
	HandleGeneric(res, 
		"type", 1,
		"file", 1,
		"object", 1,
		"action", 1,
		"status", 1,
		NULL, 0);
}

/**
 * \brief	"action" type replies handler
 *
 * Wraps HandleGeneric() with action-specific token list.
 *
 * \param	av	Session pointer
 * \param	res	Result structure pointer
 */

static void HandleAction(void *av, const struct SAVResult *res)
{
	const char *value = FindValue(res, "action", 1);
	if (!strcmp(value, "scan")) {
		HandleGeneric(res, 
			"type", 1,
			"file", 1,
			"object", 1,
			"action", 1,
			"result", 1,
			"virname", 0,
			NULL, 0);
	} else {
		HandleGeneric(res, 
			"type", 1,
			"file", 1,
			"object", 1,
			"action", 1,
			"status", 1,
			NULL, 0);
	}
}

/**
 * \brief	"ask" type replies handler
 *
 * Wraps HandleGeneric() with ask-specific token list.
 *
 * \param	av	Session pointer
 * \param	res	Result structure pointer
 */

static void HandleAsk(void *av, const struct SAVResult *res)
{
	HandleGeneric(res, 
		"type", 1,
		NULL, 0);

	printf("Leaving file\n");
	AVAnswer(av, ANSWER_LEAVE);
}

/**
 * \brief	Scan method
 *
 * Scans given path with given mode. Bails out in case of internal 
 * or sync error (file reading or scanning error is not considered 
 * internal).
 *
 * \param	av	Session pointer
 * \param	path	File or directory path
 * \param	mode	Scan mode
 *
 * \retval	0	Not infected
 * \retval	1	Infected
 */

static int Scan(void *av, const char *path, enum EAVMode mode)
{
	int done;
	int infected = 0;

	if (AVScanFile(av, path, mode) == -1)
		die("Cannot scan file: %s", strerror(errno));

	done = 0;
	while (!done) {
		struct SAVResult res;
		const char *value;

		memset(&res, 0, sizeof(res));

		if (AVResult(av, &res) == -1)
			die("Cannot get result");

		value = FindValue(&res, "type", 1);
		if (!strcmp(value, "scanner"))
			infected |= HandleScanner(av, &res, &done);
		else if (!strcmp(value, "create"))
			HandleCreate(av, &res);
		else if (!strcmp(value, "parse"))
			HandleParse(av, &res);
		else if (!strcmp(value, "action"))
			HandleAction(av, &res);
		else if (!strcmp(value, "ask"))
			HandleAsk(av, &res);
		else
			die("Invalid value \"%s\" for key \"type\"", value);

		AVResultDestroy(&res);
	}

	return infected;
}

/**
 * \brief	Prints scanner version
 *
 * \param	av	Session pointer
 * \retval	0	All ok
 * \retval	-1	Error
 */

int PrintVersion(void *av)
{
	char buf[1024];

	if (AVVersion(av, buf, sizeof(buf)) == -1)
		return -1;

	printf("%s\n", buf);
	return 0;
}

/**
 * \brief	Entry point
 *
 * Program entry point.
 *
 * \param	argc	Argument count
 * \param	argv	Argument list
 *
 * \retval	0	Not infected
 * \retval	1	Infected
 * \retval	-1	Error
 */

int main(int argc, char * const argv[])
{
	void *av;
	int infected = 0;
	struct SOptions options;
	enum EParseRet rs;
	char * const *path;

	rs = ParseOptions(&options, argc, argv);
	if (rs != PR_OK)
		return (rs == PR_EXIT_FAIL) ? EXIT_FAILURE : EXIT_SUCCESS;

	av = AVCreate(options.verbose);
	if (!av)
		die("Cannot create AV context");

	if (AVConnect(av, options.socket_path, options.reload) == -1)
		die("Cannot connect to daemon: %s", strerror(errno));

	if (PrintVersion(av) == -1)
		die("Cannot get version: %s\n", strerror(errno));

	if (options.config && AVConfig(av, options.config) == -1)
		die("Cannot send configuration: %s", strerror(errno));

	for (path = options.paths; path && *path; ++path)
		if (Scan(av, *path, options.mode))
			infected = 1;

	if (AVDisconnect(av) == -1)
		die("Cannot disconnect from daemon: %s", strerror(errno));

	AVDestroy(av);

	return infected;
}
